import logging
import re
from typing import Union

from pydantic import BaseModel

from gigax.scene import (
    Item,
    Object,
    Location,
    Character,
    ProtagonistCharacter,
    Skill,
)

logger = logging.getLogger("uvicorn")


class ActionParsingError(Exception):
    pass


class CharacterAction(BaseModel):
    command: str
    protagonist: ProtagonistCharacter
    parameters: list[Union[str, int, Object]]

    def __str__(self) -> str:
        """
        Print the action according to the training format: cmd_name param1 param2.
        e.g.: Alice: say Bob "Hello, how are you"
        """
        return f"{self.protagonist.name}: {self.command} {' '.join(map(str, self.parameters))}"

    @staticmethod
    def from_str(
        command_str: str,
        protagonist: ProtagonistCharacter,
        valid_characters: list[Character],
        valid_locations: list[Location],
        valid_items: list[Item],
        compiled_regex: re.Pattern,
    ) -> "CharacterAction":
        match = compiled_regex.match(command_str)
        if not match:
            raise ValueError("Invalid command format")

        if match.lastgroup is None:
            raise ValueError(
                f"Could not find a matching skill in command_str '{command_str}'"
            )

        command = match.lastgroup.split("_")[0]  # Extract command name
        action = CharacterAction(
            command=command, protagonist=protagonist, parameters=[]
        )

        # Extract parameters based on their named groups
        for group_name, value in match.groupdict().items():
            if not value:
                continue

            param_type = group_name[
                len(f"{command}_") :
            ]  # Remove command prefix to get parameter type

            # Add parameters based on their type
            if param_type == "character":
                # Assuming characters are directly referred by name
                action.parameters.append(
                    next(char for char in valid_characters if char.name == value),
                )
            elif param_type == "location":
                action.parameters.append(
                    next(loc for loc in valid_locations if loc.name == value),
                )
            elif param_type == "item":
                action.parameters.append(
                    next(item for item in valid_items if item.name == value)
                )
            elif param_type == "amount":
                action.parameters.append(int(value))
            elif param_type == "content":
                # Remove quotation marks if present
                action.parameters.append(value.strip('"'))

        return action


def get_guided_regex(
    skills: list[Skill],
    authorized_characters: list[Character],
    authorized_locations: list[Location],
    authorized_items: list[Item],
) -> re.Pattern:
    """
    Generate a combined regex pattern for all the skills of the protagonist.
    """
    # Get names of al authorized characters, locations, and items
    characters_names = [char.name for char in authorized_characters]
    locations_names = [loc.name for loc in authorized_locations]
    items_names = [item.name for item in authorized_items]

    # Generate a combined regex pattern for all skills
    combined_regex_parts = [
        skill.to_regex(characters_names, locations_names, items_names)
        for skill in skills
    ]
    combined_regex = "|".join(combined_regex_parts)
    return re.compile(combined_regex, re.IGNORECASE)



import pytest
from gigax.scene import Character, Item, Location, ParameterType
from gigax.parse import CharacterAction, ProtagonistCharacter, Skill


@pytest.fixture()
def context():
    # Example usage of the function
    return "A vast open world full of mystery and adventure."


@pytest.fixture()
def locations():
    return [Location(name="Old Town", description="A quiet and peaceful town.")]


@pytest.fixture()
def current_location(locations):
    return locations[0]


@pytest.fixture()
def NPCs(current_location):
    return [
        Character(
            name="John the Brave",
            description="A fearless warrior",
            current_location=current_location,
        )
    ]


@pytest.fixture()
def protagonist(current_location):
    return ProtagonistCharacter(
        name="Aldren",
        description="Brave and curious",
        current_location=current_location,
        memories=["Saved the village", "Lost a friend"],
        quests=["Find the ancient artifact", "Defeat the evil warlock"],
        skills=[
            Skill(
                name="Attack",
                description="Deliver a powerful blow",
                parameter_types=[ParameterType.character],
            )
        ],
        psychological_profile="Determined and compassionate",
    )


@pytest.fixture()
def items():
    return [Item(name="Sword", description="A sharp blade")]


@pytest.fixture()
def events(protagonist, items):
    return [
        CharacterAction(
            command="Say",
            protagonist=protagonist,
            parameters=[items[0], "What a fine sword!"],
        )
    ]



import time
import logging
import traceback

from openai import AsyncOpenAI
from gigax.prompt import NPCPrompt
from gigax.scene import (
    Character,
    Item,
    Location,
)
from dotenv import load_dotenv
from outlines import models
from outlines.generate import regex  # type: ignore
from gigax.parse import CharacterAction, ProtagonistCharacter, get_guided_regex

load_dotenv()

logger = logging.getLogger("uvicorn")


class NPCStepper:
    def __init__(
        self,
        model: str | models.LogitsGenerator,
        api_key: str | None = None,
        api_url: str = "https://gig.ax/llm/v1",
    ):
        self.model = model
        self.api_key = api_key
        self.api_url = api_url

        if isinstance(model, str) and not self.api_key:
            raise ValueError("You must provide an API key to use our API.")

    async def generate_api(
        self,
        model: str,
        prompt: str,
        guided_regex: str,
        temperature: float = 0.8,
    ) -> str:
        client = AsyncOpenAI(base_url=self.api_url, api_key=self.api_key)

        # Time the query
        start = time.time()

        messages = [
            {
                "role": "user",
                "content": prompt,
            },
        ]

        response = await client.chat.completions.create(
            model=model,
            messages=messages,
            max_tokens=100,
            temperature=temperature,
            extra_body=dict(guided_regex=guided_regex),
        )

        # Log the query time
        print(f"Query time: {time.time() - start}")

        # Return the NPC's response
        return response.choices[0].message.content  # type: ignore

    async def generate_local(
        self,
        prompt: str,
        model: models.LogitsGenerator,
        guided_regex: str,
    ) -> str:
        if not isinstance(model, (models.LlamaCpp, models.Transformers)):  # type: ignore
            raise NotImplementedError(
                "Only LlamaCpp and Transformers models are supported in local mode for now."
            )

        # Time the query
        start = time.time()

        generator = regex(model, guided_regex)
        if isinstance(model, models.LlamaCpp):  # type: ignore
            # Llama-cpp-python has a convenient create_chat_completion() method that guesses the chat prompt
            # But outlines does not support it for generation, so we do this ugly hack instead
            chat_prompt = f"<|user|>\n{prompt}<|end|>\n<|assistant|>"
        elif isinstance(model, models.Transformers):  # type: ignore
            messages = [
                {"role": "user", "content": f"{prompt}"},
            ]
            chat_prompt = model.tokenizer.tokenizer.apply_chat_template(
                messages,
                tokenize=False,
                add_generation_prompt=True,
            )
            if not isinstance(chat_prompt, str):
                raise ValueError(
                    f"Expected a string, but received type {type(chat_prompt)} with value {chat_prompt}"
                )

        res = generator(chat_prompt)
        if not isinstance(res, str):
            raise ValueError(
                f"Expected a string, but received type {type(res)} with value {res}"
            )

        logger.info(f"Query time: {time.time() - start}")
        return res

    async def get_action(
        self,
        context: str,
        locations: list[Location],
        NPCs: list[Character],
        protagonist: ProtagonistCharacter,
        items: list[Item],
        events: list[CharacterAction],
    ) -> CharacterAction:
        """
        Prompt the NPC for an input.
        """

        prompt = NPCPrompt(
            context=context,
            locations=locations,
            NPCs=NPCs,
            protagonist=protagonist,
            items=items,
            events=events,
        )

        logger.info(
            f"Prompting NPC {protagonist.name} with the following prompt: {prompt}"
        )
        guided_regex = get_guided_regex(protagonist.skills, NPCs, locations, items)

        # Generate the response
        if isinstance(self.model, models.LogitsGenerator):  # type: ignore
            res = await self.generate_local(
                prompt,
                self.model,
                guided_regex.pattern,
            )
        else:
            res = await self.generate_api(
                self.model,
                prompt,
                guided_regex.pattern,
            )

        try:
            # Parse response
            parsed_action = CharacterAction.from_str(
                res, protagonist, NPCs, locations, items, guided_regex
            )
        except Exception:
            logger.error(f"Error while parsing the action: {traceback.format_exc()}")

        logger.info(f"NPC {protagonist.name} responded with: {parsed_action}")
        return parsed_action



from gigax.parse import CharacterAction
from gigax.prompt import NPCPrompt
from gigax.scene import Character, Item, Location, ProtagonistCharacter


def test_prompt(
    context: str,
    locations: list[Location],
    NPCs: list[Character],
    protagonist: ProtagonistCharacter,
    items: list[Item],
    events: list[CharacterAction],
):
    prompt = NPCPrompt(context, locations, NPCs, protagonist, items, events)
    assert prompt

    test_prompt = """- WORLD KNOWLEDGE: A vast open world full of mystery and adventure.
- KNOWN LOCATIONS: Old Town
- NPCS: John the Brave
- CURRENT LOCATION: Old Town: A quiet and peaceful town.
- CURRENT LOCATION ITEMS: Sword
- LAST EVENTS:
Aldren: Say Sword What a fine sword!

- PROTAGONIST NAME: Aldren
- PROTAGONIST PSYCHOLOGICAL PROFILE: Brave and curious
- PROTAGONIST MEMORIES:
Saved the village
Lost a friend
- PROTAGONIST PENDING QUESTS:
Find the ancient artifact
Defeat the evil warlock
- PROTAGONIST ALLOWED ACTIONS:
Attack <character> : Deliver a powerful blow

Aldren:"""

    assert prompt == test_prompt, f"{prompt} != {test_prompt}"



from gigax.parse import CharacterAction, get_guided_regex
from gigax.scene import Character, Item, Location, ProtagonistCharacter


def test_parse(
    locations: list[Location],
    NPCs: list[Character],
    protagonist: ProtagonistCharacter,
    items: list[Item],
):
    test_command = 'attack John the Brave "Hello, how are you"'

    # Test the compiled regex
    compiled_combined_regex = get_guided_regex(
        protagonist.skills, NPCs, locations, items
    )
    assert compiled_combined_regex
    assert compiled_combined_regex.match(test_command)

    # Test the from_str method
    character_action = CharacterAction.from_str(
        test_command, protagonist, NPCs, locations, items, compiled_combined_regex
    )
    assert character_action
    assert character_action.command == "Attack"
    assert character_action.protagonist == protagonist
    assert character_action.parameters == [NPCs[0]]



import outlines
from gigax.scene import (
    Character,
    Item,
    Location,
    ProtagonistCharacter,
)
from gigax.parse import CharacterAction


@outlines.prompt
def NPCPrompt(
    context: str,
    locations: list[Location],
    NPCs: list[Character],
    protagonist: ProtagonistCharacter,
    items: list[Item],
    events: list[CharacterAction],
):
    """
    - WORLD KNOWLEDGE: {{ context }}
    - KNOWN LOCATIONS: {{ locations | map(attribute='name') | join(', ') }}
    - NPCS: {{ NPCs | map(attribute='name') | join(', ') }}
    - CURRENT LOCATION: {{ protagonist.current_location.name }}: {{ protagonist.current_location.description }}
    - CURRENT LOCATION ITEMS: {{ items | map(attribute='name') | join(', ') }}
    - LAST EVENTS:
    {% for event in events %}
    {{ event }}
    {% endfor %}

    - PROTAGONIST NAME: {{ protagonist.name }}
    - PROTAGONIST PSYCHOLOGICAL PROFILE: {{ protagonist.description }}
    - PROTAGONIST MEMORIES:
    {% for memory in protagonist.memories %}
    {{ memory }}
    {% endfor %}
    - PROTAGONIST PENDING QUESTS:
    {% for quest in protagonist.quests %}
    {{ quest }}
    {% endfor %}
    - PROTAGONIST ALLOWED ACTIONS:
    {% for skill in protagonist.skills %}
    {{ skill.to_training_format() }}
    {% endfor %}

    {{ protagonist.name }}:
    """



from enum import Enum
import re
from typing import Union
from pydantic import BaseModel, Field


class ParameterType(str, Enum):
    character = "<character>"
    location = "<location>"
    item = "<item>"
    amount = "<amount>"
    content = "<content>"
    other = "<other>"


class Object(BaseModel):
    name: str
    description: str

    def __str__(self):
        return f"{self.name}"

    def to_training_format(self) -> str:
        """
        Print the character according to the training format.
        """
        return f"{self.name}: {self.description}"


class Location(Object):
    """
    Describes a location in the game world, i.e. a town, a forest, etc.
    """

    pass


class Item(Object):
    """
    Describes an item in the game world, i.e. a sword, a potion, etc.
    """

    pass


class Character(Object):
    """
    Describes a character in the game world, i.e. an adventurer or an NPC.
    """

    current_location: Location


class Skill(BaseModel):
    """
    Model for a skill that can be performed by a character.
    e.g. "say <character> <content>", "move <location>", etc.
    """

    name: str = Field(..., description="Skill name")
    description: str = Field(..., description="Skill description")
    parameter_types: Union[list[ParameterType], dict] = (
        Field(  # This is a Union because Cubzh's Lua sends empty lists as empty dicts
            [], description="Allowed parameter types for the given skill"
        )
    )

    def to_training_format(self) -> str:
        """
        Print the action according to our training format: cmd_name <param_type1> <param_type2>
        e.g.: say <character> <content> : Say something.
        """
        return f"{self.name} {' '.join(self.parameter_types)} : {self.description}"

    def to_regex(
        self,
        character_names: list[str],
        location_names: list[str],
        item_names: list[str],
    ) -> str:
        parts = [re.escape(self.name)]
        for param in self.parameter_types:
            # Each group name follows format: skillname_paramtype, without <>
            group_name = f"{self.name}_{param.value[1:-1]}"
            if param == ParameterType.character:
                parts.append(
                    f"(?P<{group_name}>{'|'.join(map(re.escape, character_names))})"
                )
            elif param == ParameterType.location:
                parts.append(
                    f"(?P<{group_name}>{'|'.join(map(re.escape, location_names))})"
                )
            elif param == ParameterType.item:
                parts.append(
                    f"(?P<{group_name}>{'|'.join(map(re.escape, item_names))})"
                )
            elif param == ParameterType.amount:
                parts.append(f"(?P<{group_name}>\\d+)")
            elif param == ParameterType.content:
                parts.append(
                    f'(?P<{group_name}>"[^"]*")'
                )  # Match content within quotes
        return r"\s+".join(parts)


class ProtagonistCharacter(Character):
    memories: list[str] = Field(..., description="Memories that the character has.")
    quests: list[str] = Field(..., description="Quests that the character is on.")
    skills: list[Skill] = Field(..., description="Skills that the character can use.")
    psychological_profile: str



from setuptools import setup, find_packages

setup(
    name="gigax",
    version="0.1",
    description="Call LLM-powered NPCs from your game, at runtime.",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    url="http://github.com/GigaxGames/gigax",
    author="Gigax team",
    author_email="tristan@gig.ax",
    license="MIT",
    packages=find_packages(),
    install_requires=[
        "outlines",
        "pydantic",
        "openai",
        "outlines",
        "transformers",
        "llama-cpp-python",
    ],
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
    ],
    python_requires=">=3.10",
    zip_safe=False,
)



import logging
import sys


from dotenv import load_dotenv
from fastapi.exceptions import RequestValidationError
from fastapi.responses import JSONResponse
from fastapi import FastAPI, Request, status
from fastapi.middleware.cors import CORSMiddleware
from fastapi.logger import logger as fastapi_logger

from gigax.parse import CharacterAction, ProtagonistCharacter
from gigax.scene import (
    Character,
    Item,
    Location,
)
from gigax.step import NPCStepper
from prompt import NPCPrompt
from step_NPC import get_input

load_dotenv()

fastapi_logger.handlers = logging.getLogger("gunicorn.error").handlers
logging.basicConfig(level=logging.INFO, handlers=[logging.StreamHandler(sys.stdout)])

# FastAPI
app = FastAPI(root_path="/api")
origins = [
    "http://localhost",
]
app.add_middleware(
    CORSMiddleware,
    allow_origins=origins,
    allow_credentials=True,
    allow_methods=["*"],
    allow_headers=["*"],
)


logger = logging.getLogger("uvicorn")
logging.basicConfig(level=logging.INFO)


@app.exception_handler(RequestValidationError)
async def validation_exception_handler(request: Request, exc: RequestValidationError):
    exc_str = f"{exc}".replace("\n", " ").replace("   ", " ")
    logging.error(f"{request}: {exc_str}")
    content = {"status_code": 10422, "message": exc_str, "data": None}
    return JSONResponse(
        content=content, status_code=status.HTTP_422_UNPROCESSABLE_ENTITY
    )


@app.post(
    "/step",
    response_description="Step the character",
    response_model=CharacterAction,
)
async def step(
    context: str,
    locations: list[Location],
    NPCs: list[Character],
    protagonist: ProtagonistCharacter,
    current_location: Location,
    items: list[Item],
    events: list[CharacterAction],
):
    # Format the prompt
    prompt = NPCPrompt(
        context=context,
        locations=locations,
        NPCs=NPCs,
        protagonist=protagonist,
        items=items,
        events=events,
    )

    # Get the NPC's input
    # stepper = NPCStepper()


@app.get("/health-check")
async def health_check():
    return {"status": "ok"}



import os
import time
from llama_cpp import Llama
import pytest
from gigax.parse import CharacterAction
from gigax.scene import Character, Item, Location, ProtagonistCharacter
from transformers import AutoTokenizer, AutoModelForCausalLM
from gigax.step import NPCStepper
from outlines import models
from dotenv import load_dotenv

load_dotenv()


@pytest.mark.asyncio
async def test_stepper_local_llamacpp(
    context: str,
    locations: list[Location],
    NPCs: list[Character],
    protagonist: ProtagonistCharacter,
    items: list[Item],
    events: list[CharacterAction],
):
    llm = Llama(
        model_path="./models/Phi-3-mini-4k-instruct-q4.gguf",  # path to GGUF file
        n_ctx=4096,  # The max sequence length to use - note that longer sequence lengths require much more resources
        n_threads=8,  # The number of CPU threads to use, tailor to your system and the resulting performance
        n_gpu_layers=35,  # The number of layers to offload to GPU, if you have GPU acceleration available. Set to 0 if no GPU acceleration is available on your system.
    )

    model = models.LlamaCpp(llm)  # type: ignore

    stepper = NPCStepper(model=model)

    start = time.time()
    action = await stepper.get_action(
        context=context,
        locations=locations,
        NPCs=NPCs,
        protagonist=protagonist,
        items=items,
        events=events,
    )

    print(f"Query time: {time.time() - start}")
    assert str(action) == "Aldren: Attack John the Brave"


@pytest.mark.asyncio
async def test_stepper_local_transformers(
    context: str,
    locations: list[Location],
    NPCs: list[Character],
    protagonist: ProtagonistCharacter,
    items: list[Item],
    events: list[CharacterAction],
):
    llm = AutoModelForCausalLM.from_pretrained(
        "TinyLlama/TinyLlama-1.1B-Chat-v1.0", output_attentions=True
    )
    tokenizer = AutoTokenizer.from_pretrained("TinyLlama/TinyLlama-1.1B-Chat-v1.0")
    model = models.Transformers(llm, tokenizer)  # type: ignore
    # Get the NPC's input
    stepper = NPCStepper(model=model)

    action = await stepper.get_action(
        context=context,
        locations=locations,
        NPCs=NPCs,
        protagonist=protagonist,
        items=items,
        events=events,
    )

    assert str(action) == "Aldren: Attack John the Brave"


@pytest.mark.asyncio
async def test_stepper_api(
    context: str,
    locations: list[Location],
    NPCs: list[Character],
    protagonist: ProtagonistCharacter,
    items: list[Item],
    events: list[CharacterAction],
):
    # Get the NPC's input
    with pytest.raises(ValueError):
        NPCStepper(model="mistral-7b-regex")

    stepper = NPCStepper(model="mistral_7b_regex", api_key=os.getenv("API_KEY"))

    action = await stepper.get_action(
        context=context,
        locations=locations,
        NPCs=NPCs,
        protagonist=protagonist,
        items=items,
        events=events,
    )

    assert str(action) == "Aldren: Attack John the Brave"
